import numpy as np
import pandas as pd
from netneurotools import freesurfer, stats

def gen_permM(surf_files, info, hemisphere="L", n_perm=100):
    ## cortex
    coords, hemi = freesurfer.find_parcel_centroids(lhannot=surf_files[0], rhannot=surf_files[1], drop=["Unknown"])
    spins = stats.gen_spinsamples(coords=coords[hemi==0, :], hemiid=hemi[hemi==0], n_rotate=n_perm)

    ## subcortex
    info_left = info.loc[info["hemisphere"]==hemisphere, :].reset_index(drop=True)
    idx_for_permutation = info_left.loc[info_left["structure"]=="subcortex", :].index.values.copy()
    rnds = []
    for _ in range(n_perm):
        np.random.shuffle(idx_for_permutation)
        rnds.append(idx_for_permutation.copy())
    rnds = np.array(rnds).T
    return np.concatenate([spins, rnds], axis=0)  # N_region * n_perm


import numpy as np
import scipy as sp


class BNCM(object):
    def __init__(self):
        pass

    def proximal_to_distance(self, SC):
        """
        ref: Inferring neural signalling directionality from undirected structural connectomes
        """
        return -np.log( SC / (SC.max()+SC[SC>0].min()) )
    
    def calc_navigation(self, L, D, max_hops):
        """
        https://github.com/brainlife/BCT/blob/main/BCT/2019_03_03_BCT/navigation_wu.m
        """
        N = L.shape[0]
        paths = [[ [] for i in range(N) ] for j in range(N)]
        PL_bin = np.zeros(shape=(N, N))
        PL_wei = np.zeros(shape=(N, N))
        PL_dis = np.zeros(shape=(N, N))

        for i in range(N):
            for j in range(N):
                if i != j:
                    curr_node = i
                    last_node = curr_node
                    target = j
                    paths[i][j].append(curr_node)

                    pl_bin, pl_wei, pl_dis = 0, 0, 0
                    while curr_node != target:
                        neighbours_idx = np.where(~np.isnan(L[curr_node, :]))[0]
                        if len(neighbours_idx) != 0:
                            min_index = np.argmin(D[target, neighbours_idx])
                            next_node = neighbours_idx[min_index]

                        if len(neighbours_idx) == 0 or next_node == last_node or pl_bin > max_hops:
                            pl_bin = np.Inf
                            pl_wei = np.Inf
                            pl_dis = np.Inf
                            break

                        paths[i][j].append(next_node)
                        pl_bin += 1
                        pl_wei += L[curr_node, next_node]
                        pl_dis += D[curr_node, next_node]

                        last_node = curr_node
                        curr_node = next_node

                    PL_bin[i, j] = pl_bin
                    PL_wei[i, j] = pl_wei
                    PL_dis[i, j] = pl_dis
        return PL_wei
    
    def calc_communicability(W):
        degree_sqrt = np.sqrt(W.sum(axis=1, keepdims=True))
        W_norm = W / (degree_sqrt.dot(degree_sqrt.T))
        return sp.linalg.expm(W_norm)

    def calc_TransCost(self, D, G, lambda_):
        """
        ref: https://github.com/aiavenak/lambda_spectrum
        """
        D = D.copy()
        G = G.copy()

        D[np.isinf(D)] = 0
        D[np.isnan(D)] = 0
        N = D.shape[1]

        Ctrans = np.zeros(shape=(N, N))
        for t in range(N):
            Gt = G[:, [t]].T
            Gt = Gt + D
            
            FG = np.exp(-(lambda_*Gt + D)) * (D>0)
            #FG = 1 / (lambda_*Gt + D) * (D>0)
            z_lamb_t = FG.sum(axis=1, keepdims=True)
            P = FG / z_lamb_t

            idx = np.arange(N) != t
            Q = P[idx, :][:, idx].copy()
            FM = np.linalg.inv(np.eye(N-1) - Q)
            Ctrans[idx, t] = FM.dot( (P[idx, :] * D[idx, :]).sum(axis=1) )

        return Ctrans












import numpy as np
import pandas as pd
import scipy as sp
import tensorly as tl
import nibabel as nib
from statsmodels.stats import multitest

import os

def fdr_correction_ttest(Pmtx):
    idx = np.triu(np.ones_like(Pmtx), 1).astype(bool)
    p_flatten = multitest.fdrcorrection(Pmtx[idx], alpha=0.05, method="indep", is_sorted=False)
    return sp.spatial.distance.squareform(p_flatten[1])

def SC_L_decomposition(W):
    """
    ref: Decoupling of brain function from structure reveals regional behavioral specialization in humans, NC, 2019
         Maria Giulia Preti & Dimitri Van De Ville
    """
    
    degree = W.sum(axis=1)
    tmp = np.diag(degree**(-0.5))
    Wsymm = tmp.dot(W).dot(tmp)
    Wnew = Wsymm.copy()

    L = np.identity(W.shape[0]) - Wnew

    lambdaL, U = np.linalg.eigh(L)
    IndL = np.argsort(lambdaL)
    lambdaL = lambdaL[IndL]
    U = U[:, IndL]
    return lambdaL, U

class TCA(object):
    def __init__(self, A, B, C, lrdb_obj=None):
        self.A = A.copy()
        self.B = B.copy()
        self.C = C.copy()

        tmp = [self.A.sum(axis=0, keepdims=True), self.B.sum(axis=0, keepdims=True)]
        self.A /= tmp[0]
        self.B /= tmp[1]
        self.C *= (tmp[0] * tmp[1])

        self.N1 = self.A.shape[0]
        self.N2 = self.B.shape[0]
        self.n_lr = self.C.shape[0]
        self.rank = self.A.shape[1]

        if lrdb_obj is not None:
            self.L = lrdb_obj.L.copy()
            self.R = lrdb_obj.R.copy()
            self.lr_anno = lrdb_obj.lrdb.loc[lrdb_obj.TL.index, :]

            
            self.A_ = pd.DataFrame(self.A).copy()
            self.A_.index = self.L.index
            self.A_.columns = [ f"Pattern_{i+1:02d}" for i in range(self.rank) ]

            self.B_ = pd.DataFrame(self.B).copy()
            self.B_.index = self.R.index
            self.B_.columns = [ f"Pattern_{i+1:02d}" for i in range(self.rank) ]
            
            self.C_ = pd.DataFrame(self.C).copy()
            self.C_.index = self.lr_anno.index
            self.C_.columns = [ f"Pattern_{i+1:02d}" for i in range(self.rank) ]

            self.lr_anno = lrdb_obj.lrdb.loc[self.C_.index, :]
        
        self.t_stat = None
        self.t_fdr = None

    def calc_T_statistics_matrix(self):
        t_stat = np.zeros(shape=(self.N1, self.N2))
        t_pval = np.zeros(shape=(self.N1, self.N2))
        for i in range(self.N1-1):
            for j in range(i+1, self.N2):
                vec_i2j = self.C.dot( (self.A[[i], :] * self.B[[j], :]).T )
                vec_j2i = self.C.dot( (self.A[[j], :] * self.B[[i], :]).T )

                t,  p1 = sp.stats.ttest_1samp(vec_i2j-vec_j2i, 0)

                t_stat[i, j], t_pval[i, j] = t, p1
                t_stat[j, i], t_pval[j, i] = -t, p1

        t_fdr = fdr_correction_ttest(t_pval)
        self.t_stat = t_stat
        self.t_fdr = t_fdr

    def get_pathway_network(self, pathway_name, row_norm_Ce=True, return_LR_name=False):
        idx = self.lr_anno["pathway"] == pathway_name
        select_Ce = self.C_.loc[idx.values, :]
        select_name = select_Ce.index.values
        select_Ce = select_Ce.values
        if row_norm_Ce:
            select_Ce = (select_Ce - select_Ce.min(axis=1, keepdims=True)) / (select_Ce.max(axis=1, keepdims=True) - select_Ce.min(axis=1, keepdims=True))

        pathway_net = []
        for i in range(select_Ce.shape[0]):
            pathway_net.append( np.dot(self.A * select_Ce[[i], :], self.B.T) )
        pathway_net = np.array(pathway_net)

        if return_LR_name:
            return pathway_net, select_name
        else:
            return pathway_net
        pass

    def get_LR_network(self, idx, row_norm_Ce=True, return_LR_name=False):
        """Reconstruct the network of the given LR pairs

        Parameters
        ----------
        idx : list
            The index in Ce, a int or str list is required.
        return_LR_name : bool, optional
        
        Return
        ------
        lr_network : numpy.ndarray
        lr_name : numpy.ndarray, optional

        """
        ## type check
        if np.alltrue([ isinstance(i, str) for i in idx]):
            select_Ce = self.C_.loc[idx, :]
        elif np.alltrue([ isinstance(i, int) for i in idx]):
            select_Ce = self.C_.iloc[idx, :]
        else:
            raise ValueError("The data types in the list must all be str or int.")
        select_name = select_Ce.index.values
        select_Ce = select_Ce.values
        if row_norm_Ce:
            select_Ce = (select_Ce - select_Ce.min(axis=1, keepdims=True)) / (select_Ce.max(axis=1, keepdims=True) - select_Ce.min(axis=1, keepdims=True))

        
        lr_network = []
        for i in range(select_Ce.shape[0]):
            lr_network.append( np.dot(self.A * select_Ce[[i], :], self.B.T) )
        lr_network = np.array(lr_network)

        if return_LR_name:
            return lr_network, select_name
        else:
            return lr_network

        pass

    def analyze_trans_hierarchical_signaling_cortex(self, hierarchy, order="ascending",
                                                    drop_nonsignificant = False, drop_negative = True):
        """Trans-hierarchical aysmmetric signaling

        Parameters
        ----------
        hierarchy : None or numpy.ndarray
            A float array recording the hierarchical values inffered from Functional gradient, SAaxis, myelin, etc.
            If nan values in array, the corresponding region would be ignored.
            If None, Using default order of factor matrix A, by default None.
        order : str
            Sort hierarchy in "ascending" or "descending"
        drop_nonsignificant : bool, optional
            If True, the T-statistic matrix with FDR>0.05 would set to zero, by default False.
        drop_negative : bool, optional
            If True, the T-statistic matrix with values < 0 would set to zero, by default True.
        
        Return
        ------
        low2high : numpy.ndarray
            The T-statistics of lower-to-higher hierarchical regions.
        high2low : numpy.ndarray
            The T-statistics of lower-to-higher hierarchical regions.
        pval : (KS, P_KS) tuple
            The P-values of KS test.
        """
        assert self.N1 == self.N2, f"Non-square matrix doesn't support trans-hierarchial signaling analysis! ({self.N1}, {self.N2})"            
        assert order in {"ascending", "descending"}, "Error values, only 'ascending' or 'descending' are allowed."
        if self.t_stat is None:
            self.calc_T_statistics_matrix()
        
        t_stat = self.t_stat.copy()
        t_fdr = self.t_stat.copy()
        if drop_nonsignificant:
            t_stat[t_fdr>=0.05] = 0
        if drop_negative:
            t_stat[t_stat<0] = 0
        
        if hierarchy is None:
            use_region = np.array([True] * self.N1)
            use_hierarchy = np.arange(self.N1)
        else:
            use_region = ~np.isnan(hierarchy)
            use_hierarchy = hierarchy[use_region]
        use_hierarchy_idx = np.argsort(use_hierarchy)
        if order == "descending":
            use_hierarchy_idx = use_hierarchy_idx[::-1]

        t_stat = t_stat[use_region, :][:, use_region]
        t_stat = t_stat[use_hierarchy_idx, :][:, use_hierarchy_idx]
        upper_idx = np.triu(np.ones_like(t_stat) ,k=1).astype(bool)
        low2high = t_stat[upper_idx]
        high2low = t_stat[upper_idx.T]
        
        tmp_ks = sp.stats.ks_2samp(low2high, high2low, method="exact", alternative="less")  ## Null: F(x) >= G(x) Alter: F(x) < G(x)
        ks, p_ks = tmp_ks.statistic, tmp_ks.pvalue
        return low2high, high2low, (ks, p_ks)

    def analyze_trans_hierarchical_signaling_cortex_spin(self, hierarchy, order="ascending", perm_idx=None,
                                                         drop_nonsignificant = False, drop_negative = True):
        """Trans-hierarchical aysmmetric signaling

        Parameters
        ----------
        hierarchy : None or numpy.ndarray
            A float array recording the hierarchical values inffered from Functional gradient, SAaxis, myelin, etc.
            If nan values in array, the corresponding region would be ignored.
            If None, Using default order of factor matrix A, by default None.
        order : str, optional
            Sort hierarchy in "ascending" or "descending"
        perm_idx : None or (n_region, n_perm) numpy.ndarray
            The permutation index generated by random or spatial autocorrelation model.
        drop_nonsignificant : bool, optional
            If True, the T-statistic matrix with FDR>0.05 would set to zero, by default False.
        drop_negative : bool, optional
            If True, the T-statistic matrix with values < 0 would set to zero, by default True.
        
        Return
        ------
        diff_empi : float
        diff_perm : numpy.ndarray
        p_spin : float
        """
        assert self.N1 == self.N2, f"Non-square matrix doesn't support trans-hierarchial signaling analysis! ({self.N1}, {self.N2})"            
        assert order in {"ascending", "descending"}, "Error values, only 'ascending' or 'descending' are allowed."
        if self.t_stat is None:
            self.calc_T_statistics_matrix()
        
        t_stat = self.t_stat.copy()
        t_fdr = self.t_stat.copy()
        if drop_nonsignificant:
            t_stat[t_fdr>=0.05] = 0
        if drop_negative:
            t_stat[t_stat<0] = 0
        
        if hierarchy is None:
            use_region = np.array([True] * self.N1)
            use_hierarchy = np.arange(self.N1)
        else:
            use_region = ~np.isnan(hierarchy)
            use_hierarchy = hierarchy[use_region]
        use_hierarchy_idx = np.argsort(use_hierarchy)
        if order == "descending":
            use_hierarchy_idx = use_hierarchy_idx[::-1]

        t_stat = t_stat[use_region, :][:, use_region]
        t_stat = t_stat[use_hierarchy_idx, :][:, use_hierarchy_idx]
        upper_idx = np.triu(np.ones_like(t_stat) ,k=1).astype(bool)
        low2high = t_stat[upper_idx]
        high2low = t_stat[upper_idx.T]
        
        median_empi = np.median(low2high) - np.median(high2low)

        median_perm = []
        perm_idx = perm_idx[use_region, :]
        for p_idx in perm_idx.T:
            idx = np.argsort(use_hierarchy[p_idx])
            if order == "descending":
                idx = idx[::-1]
            t_perm = t_stat[idx, :][:, idx].copy()
            l2h_perm, h2l_perm = t_perm[upper_idx], t_perm[upper_idx.T]
            median_perm.append( np.median(l2h_perm)-np.median(h2l_perm) )
        median_perm = np.array(median_perm)
        p_spin = ((median_perm >= median_empi).sum()+1) / (perm_idx.shape[1]+1)
        return median_empi, median_perm, p_spin
    
    def analyze_trans_hierarchical_signaling_cortex_and_subcortex(self, hierarchy,
                                                    drop_nonsignificant = False, drop_negative = True):
        """Trans-hierarchical aysmmetric signaling

        Parameters
        ----------
        hierarchy : (numpy.ndarray, numpy.ndarry) tuple
            Two bool index recording the subcortical and cortical regions.
        drop_nonsignificant : bool, optional
            If True, the T-statistic matrix with FDR>0.05 would set to zero, by default False.
        drop_negative : bool, optional
            If True, the T-statistic matrix with values < 0 would set to zero, by default True.
        
        Return
        ------
        low2high : numpy.ndarray
            The T-statistics of lower-to-higher hierarchical regions.
        high2low : numpy.ndarray
            The T-statistics of lower-to-higher hierarchical regions.
        pval : (KS, P_KS) tuple
            The P-values of KS test.
        """
        assert self.N1 == self.N2, f"Non-square matrix doesn't support trans-hierarchial signaling analysis! ({self.N1}, {self.N2})"            
        if self.t_stat is None:
            self.calc_T_statistics_matrix()
        
        t_stat = self.t_stat.copy()
        t_fdr = self.t_stat.copy()
        if drop_nonsignificant:
            t_stat[t_fdr>=0.05] = 0
        if drop_negative:
            t_stat[t_stat<0] = 0
        
        low2high = t_stat[hierarchy[0], :][:, hierarchy[1]].ravel()
        high2low = t_stat[hierarchy[1], :][:, hierarchy[0]].ravel()
        
        tmp_ks = sp.stats.ks_2samp(low2high, high2low, method="exact", alternative="less")  ## Null: F(x) >= G(x) Alter: F(x) < G(x)
        ks, p_ks = tmp_ks.statistic, tmp_ks.pvalue
        return low2high, high2low, (ks, p_ks)

    def graph_spectrurm_analysis(self, hamonics=None, SC=None, ):
        
        ## get hamonis in ascending order
        if hamonics is None:
            if SC is None:
                raise ValueError("It shouldn't be all None for 'harmonics' and 'SC' ")
            else:
                _, hamonics = SC_L_decomposition(SC)
        
        ## get asymmetric communication for each LR pair
        Asum0, Bsum0 = self.A.sum(axis=0, keepdims=True), self.B.sum(axis=0, keepdims=True)
        sending = (self.A * Bsum0).dot(self.C.T)
        receiving = (self.B * Asum0).dot(self.C.T)
        delta = (sending - receiving).astype(np.float128)
        mean = delta.mean(axis=0, keepdims=True)
        delta = (delta - mean)

        ## decomposition
        return (np.dot(hamonics.T, delta)**2).T
    
    def identify_dominant_LR_from_Ce(self, is_norm=True, q=0.8):
        if is_norm:
            self.Ce_norm = (self.C_.T / self.C_.sum(axis=1)).T
        else:
            self.Ce_norm = self.C_.copy()
        self.Ce_bool = self.Ce_norm >= self.Ce_norm.quantile(q=q, axis=0)
    
    def prepare_ROI_from_Ae_Be(self, atlas_nii, out_path, hemi="L", thr_Ae=None, thr_Be=None):
        if thr_Ae is None:
            thr_Ae = 1 / self.rank
        if thr_Be is None:
            thr_Be = 1 / self.rank
        
        if isinstance(atlas_nii, str):
            atlas_nii = nib.load(atlas_nii)
        elif isinstance(atlas_nii, nib.nifti1.Nifti1Image):
            atlas_nii = atlas_nii
        atlas_dat = atlas_nii.get_fdata()
        
        ## processing Ae
        Ae_norm = self.A / self.A.max(axis=1, keepdims=True)
        Ae_norm = Ae_norm >= thr_Ae
        for i in range(Ae_norm.shape[1]):
            roi_label = np.argwhere(Ae_norm[:, i]).ravel()
            if hemi == "L":
                roi_label = roi_label * 2 + 1
            elif hemi == "R":
                roi_label = roi_label * 2 + 2
            else:
                pass
            
            roi_mask = np.zeros_like(atlas_dat)
            roi_mask[np.isin(atlas_dat, roi_label)] = 1
            roi_nii = nib.Nifti1Image(roi_mask, atlas_nii.affine)
            roi_nii.to_filename( os.path.join(out_path, f"Ae_pattern{i+1:02d}.nii") )
            pass
        
        ## processing Be
        Be_norm = self.B / self.B.max(axis=1, keepdims=True)
        Be_norm = Be_norm >= thr_Be
        for i in range(Be_norm.shape[1]):
            roi_label = np.argwhere(Be_norm[:, i]).ravel()
            if hemi == "L":
                roi_label = roi_label * 2 + 1
            elif hemi == "R":
                roi_label = roi_label * 2 + 2
            else:
                pass
            
            roi_mask = np.zeros_like(atlas_dat)
            roi_mask[np.isin(atlas_dat, roi_label)] = 1
            roi_nii = nib.Nifti1Image(roi_mask, atlas_nii.affine)
            roi_nii.to_filename( os.path.join(out_path, f"Be_pattern{i+1:02d}.nii") )
            pass


import numpy as np
import tensorly as tl
import os

class OBJFUNC(object):
    def __init__(self, L, R, M, TL, TR, epsilon=0, tau1=1, tau2=1, delta=1e-50, eps=1e-100,
                 device="numpy", verbose=False, log=False):

        device = self.get_device(device)
        tl.set_backend(device)
        L = tl.tensor(L, device="cuda", dtype=tl.backend.float64)
        L = tl.where(L>eps, L, eps)
        R = tl.tensor(R, device="cuda", dtype=tl.backend.float64)
        R = tl.where(R>eps, R, eps)
        if isinstance(M, tuple):
            M_type = tuple
            M1 = tl.tensor(M[0], device="cuda", dtype=tl.backend.float64)
            M2 = tl.tensor(M[1], device="cuda", dtype=tl.backend.float64)
            M = (M1, M2)
        else:
            M_type = "array"
            M = tl.tensor(M, device="cuda", dtype=tl.backend.float64)
        TL = tl.tensor(TL, device="cuda", dtype=tl.backend.float64)
        TR = tl.tensor(TR, device="cuda", dtype=tl.backend.float64)

        self.L = L
        self.R = R
        self.M = M
        self.M_type = M_type
        self.TL = TL
        self.TR = TR
        self.epsilon = epsilon
        self.tau1 = tau1
        self.tau2 = tau2
        self.delta = delta

        ##
        self.device = device
        self.verbose = verbose
        self.log = log

        ##
        self.flag_epsilon = epsilon == 0
        self.TLsum1 = TL.sum(axis=1, keepdims=True)
        self.TRsum1 = TR.sum(axis=1, keepdims=True)
        self.const_T = self.tau1 * self.TLsum1 + self.tau2 * self.TRsum1
        self.logL = tl.log(L)
        self.logR = tl.log(R)
        self.Lsum = tau1 * self.L.sum()
        self.Rsum = tau2 * self.R.sum()
        pass

    def f(self, A, B, C):
        Csum0 = C.sum(axis=0, keepdims=True)
        #print(self.M.shape, A.shape, B.shape, C.shape)
        if self.M_type == "array":
            return (self.M * tl.matmul(A*Csum0, B.T)).sum()
        else:
            return ( tl.matmul(self.M[1].T, B) * tl.matmul(self.M[0].T, A*Csum0) ).sum()
    def grad_f(self, A, B, C):
        Csum0 = C.sum(axis=0, keepdims=True)
        if self.M_type == "array":
            grad_A = tl.matmul(self.M, B*Csum0)
            grad_B = tl.matmul(self.M.T, A*Csum0)
            grad_C = (A*tl.matmul(self.M, B)).sum(axis=0, keepdims=True)
        else:
            grad_A = tl.matmul(self.M[0], tl.matmul(self.M[1].T, B*Csum0))
            grad_B = tl.matmul(self.M[1], tl.matmul(self.M[0].T, A*Csum0))
            grad_C = (tl.matmul(self.M[0].T, A) * tl.matmul(self.M[1].T, B)).sum(axis=0, keepdims=True)
        #if self.verbose:
        #    print(grad_A.shape, grad_B.shape, grad_C.shape)
        return [grad_A, grad_B, grad_C]
    def grad_fA(self, A, B, C):
        Csum0 = C.sum(axis=0, keepdims=True)
        if self.M_type == "array":
            return tl.matmul(self.M, B*Csum0)
        else:
            return tl.matmul(self.M[0], tl.matmul(self.M[1].T, B*Csum0))
    def grad_fB(self, A, B, C):
        Csum0 = C.sum(axis=0, keepdims=True)
        if self.M_type == "array":
            return tl.matmul(self.M.T, A*Csum0)
        else:
            return tl.matmul(self.M[1], tl.matmul(self.M[0].T, A*Csum0))
    def grad_fC(self, A, B, C):
        if self.M_type == "array":
            return (A*tl.matmul(self.M, B)).sum(axis=0, keepdims=True)
        else:
            return (tl.matmul(self.M[0].T, A) * tl.matmul(self.M[1].T, B)).sum(axis=0, keepdims=True)
    
    def e(self, A, B, C):
        if self.flag_epsilon:
            return 0
        else:
            return self.epsilon * sum([  self.nega_entropy(tmp) for tmp in (A, B, C) ])
    def grad_e(self, A, B, C):
        if self.flag_epsilon:
            return [0, 0, 0]
        else:
            return [ tl.log(tmp) * self.epsilon for tmp in (A, B, C)]
    
    def tau1KL(self, A, B, C):
        Bsum0 = B.sum(axis=0, keepdims=True)
        H1 = tl.matmul( A*Bsum0, tl.matmul(C.T, self.TL) )
        return self.tau1 * self.kl_div(H1, self.logL) + self.Lsum
    def grad_tau1KL(self, A, B, C):
        Bsum0 = B.sum(axis=0, keepdims=True)
        TLC = tl.matmul(self.TL.T, C)
        D1 = TLC * Bsum0
        H1 = tl.matmul(A, D1.T)
        grad_A = tl.matmul( (tl.log(H1)-tl.log(self.L)), D1)
        grad_B = ( TLC * tl.matmul((tl.log(H1)-tl.log(self.L)).T, A) ).sum(axis=0, keepdims=True)
        grad_C = tl.matmul(self.TL, tl.matmul((tl.log(H1)-tl.log(self.L)).T, A*Bsum0))
        if self.verbose:
            print(grad_A.shape, grad_B.shape, grad_C.shape)
        return [ tmp*self.tau1 for tmp in (grad_A, grad_B, grad_C)]
    
    def tau2KL(self, A, B, C):
        Asum0 = A.sum(axis=0, keepdims=True)
        H2 = tl.matmul( B*Asum0, tl.matmul(C.T, self.TR) )
        return self.tau2 * self.kl_div(H2, self.logR) + self.Rsum
    def grad_tau2KL(self, A, B, C):
        Asum0 = A.sum(axis=0, keepdims=True)
        TRC = tl.matmul(self.TR.T, C)
        D2 = TRC * Asum0
        H2 = tl.matmul(B, D2.T)
        grad_A = ( TRC * tl.matmul((tl.log(H2)-tl.log(self.R)).T, B) ).sum(axis=0, keepdims=True)
        grad_B = self.tau2 * tl.matmul( (tl.log(H2)-tl.log(self.R)), D2)
        grad_C = tl.matmul(self.TR, tl.matmul((tl.log(H2)-tl.log(self.R)).T, B*Asum0))
        if self.verbose:
            print(grad_A.shape, grad_B.shape, grad_C.shape)
        return [ tmp*self.tau1 for tmp in (grad_A, grad_B, grad_C)]

    def uAL(self, A, A_, B_, C_):
        Bsum0 = B_.sum(axis=0, keepdims=True)
        D1 = tl.matmul(self.TL.T, C_) * Bsum0
        D1sum0 = D1.sum(axis=0, keepdims=True)
        H1 = tl.matmul(A, D1.T)
        H1_ = tl.matmul(A_, D1.T)
        res = ((tl.log(A) - tl.log(A_))*A*D1sum0).sum() \
            + (H1*(tl.log(H1_)-self.logL-1)).sum()
        return res * self.tau1 + self.Lsum
    def grad_uAL(self, A, A_, B_, C_):
        Bsum0 = B_.sum(axis=0, keepdims=True)
        D1 = tl.matmul(self.TL.T, C_) * Bsum0
        D1sum0 = D1.sum(axis=0, keepdims=True)
        H1_ = tl.matmul(A_, D1.T)
        grad_A = tl.log(A/A_)*D1sum0 \
            + tl.matmul(tl.log(H1_)-self.logL, D1)
        return self.tau1 * grad_A
    def min_uAL(self, A_, B_, C_):
        Bsum0 = B_.sum(axis=0, keepdims=True)
        D1 = tl.matmul(self.TL.T, C_) * Bsum0
        D1sum0 = D1.sum(axis=0, keepdims=True)
        H1_ = tl.matmul(A_, D1.T)
        return A_ * tl.exp( tl.matmul(self.logL-tl.log(H1_), D1)/D1sum0 )
    
    def uAR(self, A, A_, B_, C_):
        Asum0 = A.sum(axis=0, keepdims=True)
        A_sum0 = A_.sum(axis=0, keepdims=True)
        Bsum0 = B_.sum(axis=0, keepdims=True)
        CR = tl.matmul(C_.T, self.TR)
        CRsum1 = CR.sum(axis=1)
        H2 = tl.matmul(B_*Asum0, CR)
        H2_ = tl.matmul(B_*A_sum0, CR)
        res = ((tl.log(A)-tl.log(A_))*(A*(Bsum0*CRsum1))).sum() \
            + (H2*(tl.log(H2_)-self.logR-1)).sum()
        return self.tau2 * res + self.Rsum
    def grad_uAR(self, A, A_, B_, C_):
        A_sum0 = A_.sum(axis=0, keepdims=True)
        Bsum0 = B_.sum(axis=0, keepdims=True)
        CR = tl.matmul(C_.T, self.TR)
        CRsum1 = CR.sum(axis=1)
        H2_ = tl.matmul(B_*A_sum0, CR)
        grad_A = tl.log(A/A_)*(Bsum0*CRsum1) \
               + (CR.T * tl.matmul((tl.log(H2_)-self.logR).T, B_)).sum(axis=0, keepdims=True)
        return self.tau2 * grad_A
    def min_uAR(self, A_, B_, C_):
        A_sum0 = A_.sum(axis=0, keepdims=True)
        Bsum0 = B_.sum(axis=0, keepdims=True)
        CR = tl.matmul(C_.T, self.TR)
        CRsum1 = CR.sum(axis=1)
        H2_ = tl.matmul(B_*A_sum0, CR)
        omega = (CR.T * tl.matmul((tl.log(H2_)-self.logR).T, B_)).sum(axis=0, keepdims=True)
        lamda = Bsum0*CRsum1
        return A_ * tl.exp(-omega/lamda)

    def uBL(self, B, B_, A_, C_):
        Asum0 = A_.sum(axis=0, keepdims=True)
        Bsum0 = B.sum(axis=0, keepdims=True)
        B_sum0 = B_.sum(axis=0, keepdims=True)
        CL = tl.matmul(C_.T, self.TL)
        CLsum1 = CL.sum(axis=1)
        H1 = tl.matmul(A_*Bsum0, CL)
        H1_ = tl.matmul(A_*B_sum0, CL)
        res = ((tl.log(B)-tl.log(B_))*(B*(Asum0*CLsum1))).sum() \
            + (H1*(tl.log(H1_)-self.logL-1)).sum()
        return self.tau1 * res + self.Lsum
    def grad_uBL(self, B, B_, A_, C_):
        Asum0 = A_.sum(axis=0, keepdims=True)
        B_sum0 = B_.sum(axis=0, keepdims=True)
        CL = tl.matmul(C_.T, self.TL)
        CLsum1 = CL.sum(axis=1)
        H1_ = tl.matmul(A_*B_sum0, CL)
        grad_B = tl.log(B/B_)*(Asum0*CLsum1) \
               + (CL.T * tl.matmul((tl.log(H1_)-self.logL).T, A_)).sum(axis=0, keepdims=True)
        return self.tau1 * grad_B
    def min_uBL(self, B_, A_, C_):
        Asum0 = A_.sum(axis=0, keepdims=True)
        B_sum0 = B_.sum(axis=0, keepdims=True)
        CL = tl.matmul(C_.T, self.TL)
        CLsum1 = CL.sum(axis=1)
        H1_ = tl.matmul(A_*B_sum0, CL)
        omega = (CL.T * tl.matmul((tl.log(H1_)-self.logL).T, A_)).sum(axis=0, keepdims=True)
        lamda = Asum0*CLsum1
        return B_ * tl.exp(-omega/lamda)

        pass

    def uBR(self, B, B_, A_, C_):
        Asum0 = A_.sum(axis=0, keepdims=True)
        D2 = tl.matmul(self.TR.T, C_) * Asum0
        D2sum0 = D2.sum(axis=0, keepdims=True)
        H2 = tl.matmul(B, D2.T)
        H2_ = tl.matmul(B_, D2.T)
        res = ((tl.log(B) - tl.log(B_))*B*D2sum0).sum() \
            + (H2*(tl.log(H2_)-self.logR-1)).sum()
        return res * self.tau2 + self.Rsum
    def grad_uBR(self, B, B_, A_, C_):
        Asum0 = A_.sum(axis=0, keepdims=True)
        D2 = tl.matmul(self.TR.T, C_) * Asum0
        D2sum0 = D2.sum(axis=0, keepdims=True)
        H2_ = tl.matmul(B_, D2.T)
        grad_A = tl.log(B/B_)*D2sum0 \
            + tl.matmul(tl.log(H2_)-self.logR, D2)
        return self.tau1 * grad_A
    def min_uBR(self, B_, A_, C_):
        Asum0 = A_.sum(axis=0, keepdims=True)
        D2 = tl.matmul(self.TR.T, C_) * Asum0
        D2sum0 = D2.sum(axis=0, keepdims=True)
        H2_ = tl.matmul(B_, D2.T)
        return B_ * tl.exp( tl.matmul(self.logR-tl.log(H2_), D2)/D2sum0 )


    def uCL(self, C, C_, A_, B_):
        Asum0 = A_.sum(axis=0, keepdims=True)
        Bsum0 = B_.sum(axis=0, keepdims=True)
        CL = tl.matmul(C.T, self.TL)
        CL_ = tl.matmul(C_.T, self.TL)
        H1 = tl.matmul(A_*Bsum0, CL)
        H1_ = tl.matmul(A_*Bsum0, CL_)
        res = (tl.log(C/C_)*C*self.TLsum1*Asum0*Bsum0).sum() \
            + (H1*(tl.log(H1_)-self.logL-1)).sum()
        return self.tau1*res + self.Lsum
    def grad_uCL(self, C, C_, A_, B_):
        Asum0 = A_.sum(axis=0, keepdims=True)
        Bsum0 = B_.sum(axis=0, keepdims=True)
        CL_ = tl.matmul(C_.T, self.TL)
        H1_ = tl.matmul(A_*Bsum0, CL_)
        tmp = tl.matmul( (tl.log(H1_)-self.logL).T, A_*Bsum0)
        grad_C = (tl.log(C/C_)*self.TLsum1*Asum0*Bsum0) \
               + tl.matmul(self.TL, tmp)
        return self.tau1 * grad_C
    def min_uCL(self, C_, A_, B_):
        Asum0 = A_.sum(axis=0, keepdims=True)
        Bsum0 = B_.sum(axis=0, keepdims=True)
        CL_ = tl.matmul(C_.T, self.TL)
        H1_ = tl.matmul(A_*Bsum0, CL_)
        tmp = tl.matmul( (tl.log(H1_)-self.logL).T, A_*Bsum0)
        omega = tl.matmul(self.TL, tmp)
        lamda = self.TLsum1 * Asum0 * Bsum0
        return C_ * tl.exp(-omega/lamda)
    
    def uCR(self, C, C_, A_, B_):
        Asum0 = A_.sum(axis=0, keepdims=True)
        Bsum0 = B_.sum(axis=0, keepdims=True)
        CR = tl.matmul(C.T, self.TR)
        CR_ = tl.matmul(C_.T, self.TR)
        H2 = tl.matmul(B_*Asum0, CR)
        H2_ = tl.matmul(B_*Asum0, CR_)
        res = (tl.log(C/C_)*C*self.TRsum1*Asum0*Bsum0).sum() \
            + (H2*(tl.log(H2_)-self.logR-1)).sum()
        return self.tau2*res + self.Rsum
    def grad_uCR(self, C, C_, A_, B_):
        Asum0 = A_.sum(axis=0, keepdims=True)
        Bsum0 = B_.sum(axis=0, keepdims=True)
        CR_ = tl.matmul(C_.T, self.TR)
        H2_ = tl.matmul(B_*Asum0, CR_)
        tmp = tl.matmul( (tl.log(H2_)-self.logR).T, B_*Asum0)
        grad_C = (tl.log(C/C_)*self.TRsum1*Asum0*Bsum0) \
               + tl.matmul(self.TR, tmp)
        return self.tau2 * grad_C
    def min_uCR(self, C_, A_, B_):
        Asum0 = A_.sum(axis=0, keepdims=True)
        Bsum0 = B_.sum(axis=0, keepdims=True)
        CR_ = tl.matmul(C_.T, self.TR)
        H2_ = tl.matmul(B_*Asum0, CR_)
        tmp = tl.matmul( (tl.log(H2_)-self.logR).T, B_*Asum0)
        omega = tl.matmul(self.TR, tmp)
        lamda = self.TRsum1 * Asum0 * Bsum0
        return C_ * tl.exp(-omega/lamda)

    def min_uCL_UCR(self, C_, A_, B_):
        Asum0 = A_.sum(axis=0, keepdims=True)
        Bsum0 = B_.sum(axis=0, keepdims=True)
        
        CL_ = tl.matmul(C_.T, self.TL)
        H1_ = tl.matmul(A_*Bsum0, CL_)
        tmp1 = tl.matmul( (tl.log(H1_)-self.logL).T, A_*Bsum0)

        CR_ = tl.matmul(C_.T, self.TR)
        H2_ = tl.matmul(B_*Asum0, CR_)
        tmp2 = tl.matmul( (tl.log(H2_)-self.logR).T, B_*Asum0)

        omega = self.tau1 * tl.matmul(self.TL, tmp1) + self.tau2 * tl.matmul(self.TR, tmp2)
        lamda = self.const_T * Asum0 * Bsum0
        return C_ * tl.exp(-omega/lamda)


    def F(self, A, B, C):
        return self.f(A, B, C) + self.e(A, B, C) + self.tau1KL(A, B, C) + self.tau2KL(A, B, C)
    def grad_F(self, A, B, C):
        grad_f = self.grad_f(A, B, C)
        grad_e = self.grad_e(A, B, C)
        grad_tau1KL = self.grad_tau1KL(A, B, C)
        grad_tau2KL = self.grad_tau2KL(A, B, C)
        return [grad_f[i]+grad_e[i]+grad_tau1KL[i]+grad_tau2KL[i] for i in range(3)]
    
    def GA(self, A, A_, B_, C_):
        return self.f(A, B_, C_) + self.e(A, B_, C_) \
              + self.uAL(A, A_, B_, C_) + self.uAR(A, A_, B_, C_)
    def grad_GA(self, A, A_, B_, C_):
        return self.grad_fA(A, B_, C_) + self.grad_e(A, B_, C_)[0] \
              + self.grad_uAL(A, A_, B_, C_) + self.grad_uAR(A, A_, B_, C_)
    def min_GA(self, A_, B_, C_):
        ## grad_f
        grad_fA = self.grad_fA(A_, B_, C_)
        
        ## grad_uAL
        Bsum0 = B_.sum(axis=0, keepdims=True)
        D1 = tl.matmul(self.TL.T, C_) * Bsum0
        D1sum0 = D1.sum(axis=0, keepdims=True)
        H1_ = tl.matmul(A_, D1.T)
        
        ## grad_uAR
        A_sum0 = A_.sum(axis=0, keepdims=True)
        #Bsum0 = B_.sum(axis=0, keepdims=True)
        CR = tl.matmul(C_.T, self.TR)
        CRsum1 = CR.sum(axis=1)
        H2_ = tl.matmul(B_*A_sum0, CR)

        omega = grad_fA \
              + self.tau1 * tl.matmul(tl.log(H1_)-self.logL, D1) \
              + self.tau2 * (CR.T * tl.matmul((tl.log(H2_)-self.logR).T, B_)).sum(axis=0, keepdims=True)
        lamda = D1sum0 + (Bsum0*CRsum1)
        if self.flag_epsilon:
            return A_ * tl.exp(-omega/lamda)
        else:
            scale = lamda + self.epsilon
            return (A_**(lamda/scale)) * tl.exp(-omega/scale)

    def GB(self, B, B_, A_, C_):
        return self.f(A_, B, C_) + self.e(A_, B, C_) \
              + self.uBL(B, B_, A_, C_) + self.uBR(B, B_, A_, C_)
    def grad_GB(self, B, B_, A_, C_):
        return self.grad_fB(A_, B, C_) + self.grad_e(A_, B, C_)[1] \
              + self.grad_uBL(B, B_, A_, C_) + self.grad_uBR(B, B_, A_, C_)
    def min_GB(self, B_, A_, C_):
        ## grad_f
        grad_fB = self.grad_fB(A_, B_, C_)

        ## grad_uBL
        Asum0 = A_.sum(axis=0, keepdims=True)
        B_sum0 = B_.sum(axis=0, keepdims=True)
        CL = tl.matmul(C_.T, self.TL)
        CLsum1 = CL.sum(axis=1)
        H1_ = tl.matmul(A_*B_sum0, CL)

        ## grad_uBR
        #Asum0 = A_.sum(axis=0, keepdims=True)
        D2 = tl.matmul(self.TR.T, C_) * Asum0
        D2sum0 = D2.sum(axis=0, keepdims=True)
        H2_ = tl.matmul(B_, D2.T)
        omega = grad_fB \
              + self.tau1 * (CL.T * tl.matmul((tl.log(H1_)-self.logL).T, A_)).sum(axis=0, keepdims=True) \
              + self.tau2 * tl.matmul(tl.log(H2_)-self.logR, D2)
        lamda = (Asum0*CLsum1) + D2sum0
        if self.flag_epsilon:
            return B_ * tl.exp(-omega/lamda)
        else:
            scale = lamda + self.epsilon
            return (B_**(lamda/scale)) * tl.exp(-omega/scale)

    def GC(self, C, C_, A_, B_):
        return self.f(A_, B_, C) + self.e(A_, B_, C) \
              + self.uCL(C, C_, A_, B_) + self.uCR(C, C_, A_, B_)
    def grad_GC(self, C, C_, A_, B_):
        return self.grad_fC(A_, B_, C) + self.grad_e(A_, B_, C) \
              + self.grad_uCL(C, C_, A_, B_) + self.grad_uCR(C, C_, A_, B_)
    def min_GC(self, C_, A_, B_):
        ## grad_f
        grad_fC = self.grad_fC(A_, B_, C_)

        ## grad_uCL
        Asum0 = A_.sum(axis=0, keepdims=True)
        Bsum0 = B_.sum(axis=0, keepdims=True)
        CL_ = tl.matmul(C_.T, self.TL)
        H1_ = tl.matmul(A_*Bsum0, CL_)
        tmp1 = tl.matmul( (tl.log(H1_)-self.logL).T, A_*Bsum0)

        ## grad_uCR
        #Asum0 = A_.sum(axis=0, keepdims=True)
        #Bsum0 = B_.sum(axis=0, keepdims=True)
        CR_ = tl.matmul(C_.T, self.TR)
        H2_ = tl.matmul(B_*Asum0, CR_)
        tmp2 = tl.matmul( (tl.log(H2_)-self.logR).T, B_*Asum0)
        
        omega = grad_fC \
              + self.tau1 * tl.matmul(self.TL, tmp1) \
              + self.tau2 * tl.matmul(self.TR, tmp2)
        lamda = tl.matmul(self.const_T, Asum0*Bsum0)
        if self.flag_epsilon:
            return C_ * tl.exp(-omega/lamda)
        else:
            scale = lamda + self.epsilon
            return (C_**(lamda/scale)) * tl.exp(-omega/scale)

    def nega_entropy(self, x):
        return (x*(tl.log(x)-1)).sum()
    
    def kl_div(self, x, logy):
        return (x*(tl.log(x)-logy-1)).sum()

    def get_device(self, device):
        if device == "numpy":
            device = "numpy"
        elif device == "pytorch":
            try:
                import torch
                if torch.cuda.is_available():
                    device = "pytorch"
                else:
                    device = "numpy"
                    print("GPU is not avaliable, use numpy as default.")
            except:
                device = "numpy"
        else:
            raise ValueError("Only numpy or pytorch is avaliable.")
        return device

class MMOTNTD(OBJFUNC):
    """MMOTNTD

    Using BMM algorithm to solve the following problem:
     (A^*, B^*, C^*) = argmin <M, Adiag(C^T1)B^T>_F + \epsilon H(A,B,C) 
                          + \tau_1 KL(Adiag(B^T1)C^TT_L) + \tau_2 KL(Bdiag(A^T1)C^TT_R)
                    s.t. A >= \delta, B >= \delta, C >= \delta

    Parameters
    ----------
    L : (n_sender, n_ligand) numpy.ndarray
        The ligand expression matrix.
    R : (n_receiver, n_receptor) numpy.ndarray
        The receptor expression matrix.
    M : (n_sender, n_receiver) numpy.ndarray or tuple (M1, M2)
        The transport cost matrix computed using inverse optimal transport.
        If the input type is tuple, it should include two component that decomposed 
        from transport cost matrxi M.
    TL : (n_lr, n_ligand) numpy.ndarray
        The generated ligand coupling matrix (a binary matrix).
    TR : (n_lr, n_receptor) numpy.ndarray
        The generated ligand coupling matrix (a binary matrix).
    epsilon : int or float
        The coefficient of entropy regularization of tensor factor A, B, C, by default 0.
    tau1 : int or float
        The coefficient for KL divergence of unmatched ligand expression, by default 1.
    tau2 : int or float
        The coefficient for KL divergence of unmatched receptor expression, by default 1.
    delta : float
        The lower bound of tensor factor A, B, C, by default 1e-100.
    eps : float
        The lower bound of ligand or receptor expression, by default 1e-80.
    verbose : bool
        Whether to display the result of each iteration, by default False.
    log : bool
        Whether to save the result of each iteration, by default False.
    device : str, "numpy" or "pytorch"
        The backend used to perform this algorithm.

    Attributes
    ----------
    factors: tuple
        The solved result (A^*, B^*, C^*)
    
    """
    def __init__(self, L, R, M, TL, TR, epsilon=0, tau1=1, tau2=1, 
                 delta=1e-100, eps=1e-80, 
                 verbose=False, log=False, device="numpy"):
        super(MMOTNTD, self).__init__(L, R, M, TL, TR, epsilon, tau1, tau2, delta, eps,
                                       device, verbose, log)
    
    def random_init1(self, r, n_iter):
        N1, I_len = self.L.shape
        N2 = self.R.shape[0]
        if self.TL is not None:
            I_len = self.TL.shape[0]

        A0 = tl.random.random_tensor(shape=(N1, r))
        A0 = tl.where(A0>self.delta, A0, self.delta)
        B0 = tl.random.random_tensor(shape=(N2, r))
        B0 = tl.where(B0>self.delta, B0, self.delta)
        C0 = tl.random.random_tensor(shape=(I_len, r))
        C0 = tl.where(C0>self.delta, C0, self.delta)
        for _ in range(n_iter):
            Amin = self.min_uAR(A0, B0, C0)
            Bmin = self.min_uBL(B0, A0, C0)
            Cmin = (self.min_uCL(C0, A0, B0) + self.min_uCR(C0, A0, B0))/2

            A0 = tl.where(Amin>self.delta, Amin, self.delta)
            B0 = tl.where(Bmin>self.delta, Bmin, self.delta)
            C0 = tl.where(Cmin>self.delta, Cmin, self.delta)
        return A0, B0, C0
    
    def random_init2(self, r, n_iter1=10, n_iter2=50, n_round=100):
        if n_round == 0:
             optA, optB, optC = self.random_init1(r, n_iter=n_iter1)
        else:
            optF = np.Inf
            for _ in range(n_round):
                A_test, B_test, C_test = self.random_init1(r, n_iter=n_iter1)
                for _ in range(n_iter2):
                    C_test = self.min_GC(C_test, A_test, B_test)
                    C_test = tl.where(C_test>self.delta, C_test, self.delta)
                    A_test = self.min_GA(A_test, B_test, C_test)
                    A_test = tl.where(A_test>self.delta, A_test, self.delta)
                    B_test = self.min_GB(B_test, A_test, C_test)
                    B_test = tl.where(B_test>self.delta, B_test, self.delta)
                curF = self.F(A_test, B_test, C_test)
                if optF > curF:
                    optF = curF
                    optA, optB, optC = A_test.copy(), B_test.copy(), C_test.copy()
        return optA, optB, optC

    def get_device(self, device):
        if device == "numpy":
            device = "numpy"
        elif device == "pytorch":
            try:
                import torch
                if torch.cuda.is_available():
                    device = "pytorch"
                else:
                    device = "numpy"
                    print("GPU is not avaliable, use numpy as default.")
            except:
                device = "numpy"
        else:
            raise ValueError("Only numpy or pytorch is avaliable.")
        return device 

    def fit(self, r, 
            init_val = None, init_params = [5, 50, 100],   ##KL, OT, n_round,
            block_order = "sequential", # or "random"
            min_iter=300, max_iter = 5000, stopThr = [1e-8, 1e-5], n_cpu = None,
            verbose=False, log=False):
        """

        Parameters
        ----------
        r : int
            Number of input rank
        init_val : None or tuple (A, B, C), optional
            The initial value of tensor factor A, B, C to iteration, by default None
        init_params : list[num1, num2, num3], optional
            The first number is the iteration step for KL divergence. The second numbder is the 
            iteration step for MMOTNTD. The third number is the round for repeat the above procedure.
            Set [0, 0, 0] for random initialization. By default [5, 50, 100]
        block_order : str "sequential" or "random", optional
            If "sequential", the order of block to update is C, B, A.
            If "random",randomize the order of coordinates in the BMM solver. 
            By default "sequential".
        min_iter : int, optional
            Minimum number of iterations, by default 300
        max_iter : int, optional
            Maximum number of iterations before timing out, by default 5000
        stopThr : list[num1, num2], optional
            The first number is absolute error of result between adjcent iteration.
            The second number is relative error of result between adjcent iteration.
            By default [1e-8, 1e-5].
        n_cpu : None or int, optional
            The number cpus to perform the iteration, by default None
        verbose : bool
            Whether to display the result of each iteration, by default False.
        log : bool
            Whether to save the result of each iteration, by default False.
        """
        
        ## reset parameters
        self.min_iter = min_iter
        self.max_iter = max_iter
        self.stopThr = stopThr
        
        self.verbose = verbose
        self.log = log

        ## set n_cpu
        tl.set_backend(self.device)
        if self.device == "numpy" and n_cpu is not None:
            os.environ["OMP_NUM_THREADS"] = str(n_cpu)            # export OMP_NUM_THREADS=4
            os.environ["OPENBLAS_NUM_THREADS"] = str(n_cpu)       # export OPENBLAS_NUM_THREADS=4
            os.environ["MKL_NUM_THREADS"] = str(n_cpu)            # export MKL_NUM_THREADS=6
            os.environ["VECLIB_MAXIMUM_THREADS"] = str(n_cpu)     # export VECLIB_MAXIMUM_THREADS=4
            os.environ["NUMEXPR_NUM_THREADS"] = str(n_cpu)
        pass
        
        ## Initiation
        if init_val is None:
            A, B, C = self.random_init2(r, n_iter1=init_params[0], n_iter2=init_params[1], n_round=init_params[2])
        else:
            A = tl.tensor(init_val[0], device="cuda", dtype=tl.backend.float64)
            B = tl.tensor(init_val[1], device="cuda", dtype=tl.backend.float64)
            C = tl.tensor(init_val[2], device="cuda", dtype=tl.backend.float64)
        
        ## Update
        if block_order == "sequential":
            (Ae, Be, Ce) = self.sequential_update(A, B, C)
        elif block_order == "random":
            (Ae, Be, Ce) = self.random_update(A, B, C)

        self.factors = (Ae, Be, Ce)
    
    def rescale_tensor_factor(self, A, B, C):
        tmp = [A.sum(axis=0, keepdims=True), B.sum(axis=0, keepdims=True)]
        return A/tmp[0], B/tmp[1], C * (tmp[0] * tmp[1])

    def calc_obj_res(self, A, B, C):
        res = [ self.f(A,B,C), self.e(A, B, C), self.tau1KL(A, B, C), self.tau2KL(A, B, C) ]
        res.append(sum(res[:4]))
        return res

    def calc_var_err(self, var_prev, var_curr, method="max"):
        if method == "max":
            return [ tl.max(tl.abs(a-b)) for a, b in zip(var_prev, var_curr)]
        elif method == "norm2":
            return [ tl.norm(a-b, order=2) for a, b in zip(var_prev, var_curr)]
    
    def sequential_update(self, A0, B0, C0):
        A_test, B_test, C_test = A0.copy(), B0.copy(), C0.copy()
        prev_obj_val = self.F(A_test, B_test, C_test)
        
        if self.log:
            self.logmm = []
        
        N_continue, last_k = 0, 0
        for k in range(self.max_iter):
            #A_prev, B_prev, C_prev = A_test.copy(), B_test.copy(), C_test.copy()
            C_test = self.min_GC(C_test, A_test, B_test)
            C_test = tl.where(C_test>self.delta, C_test, self.delta)
            A_test = self.min_GA(A_test, B_test, C_test)
            A_test = tl.where(A_test>self.delta, A_test, self.delta)
            B_test = self.min_GB(B_test, A_test, C_test)
            B_test = tl.where(B_test>self.delta, B_test, self.delta)

            curr_obj_val = self.F(A_test, B_test, C_test)
            err = tl.abs(curr_obj_val-prev_obj_val)
            rel_err = err / tl.abs(curr_obj_val)
            #var_err1 = self.calc_var_err( (A_prev, B_prev, C_prev), (A_test, B_test, C_test), method="max")
            #var_err2 = self.calc_var_err( (A_prev, B_prev, C_prev), (A_test, B_test, C_test), method="norm2")
            prev_obj_val = curr_obj_val
            
            if self.log:
                self.logmm.append([curr_obj_val, rel_err, err])
            if self.verbose:
                print(k, curr_obj_val, rel_err, err, N_continue)

            if rel_err <= self.stopThr[0] and err <= self.stopThr[1]:
                if (k - last_k) == 1:
                    N_continue += 1
                else:
                    N_continue = 1
                last_k = k
            else:
                N_continue = 0
            if k>self.min_iter and N_continue >= 10:
                self.n_total_iter = k
                break
        self.n_total_iter = k
        return (A_test, B_test, C_test)
    
    def random_update(self, A0, B0, C0):
        A_test, B_test, C_test = A0.copy(), B0.copy(), C0.copy()
        prev_obj_val = self.F(A_test, B_test, C_test)

        if self.log:
            self.logmm = []

        N_continue, last_k = 0, 0
        n_block = 3
        #for k, idx in enumerate(np.random.randint(0, 3, size=self.max_iter*n_block)):
        order = np.array(range(n_block))
        for k in range(self.max_iter):
            np.random.shuffle(order)
            #A_prev, B_prev, C_prev = A_test.copy(), B_test.copy(), C_test.copy()
            for idx in order:
                if idx == 0:
                    A_test = self.min_GA(A_test, B_test, C_test)
                    A_test = tl.where(A_test>self.delta, A_test, self.delta)
                elif idx == 1:
                    B_test = self.min_GB(B_test, A_test, C_test)
                    B_test = tl.where(B_test>self.delta, B_test, self.delta)
                elif idx == 2:
                    C_test = self.min_GC(C_test, A_test, B_test)
                    C_test = tl.where(C_test>self.delta, C_test, self.delta)                        

            curr_obj_val = self.F(A_test, B_test, C_test)
            err = tl.abs(curr_obj_val-prev_obj_val)
            rel_err = err / tl.abs(curr_obj_val)
            #var_err1 = self.calc_var_err( (A_prev, B_prev, C_prev), (A_test, B_test, C_test), method="max")
            #var_err2 = self.calc_var_err( (A_prev, B_prev, C_prev), (A_test, B_test, C_test), method="norm2")

            prev_obj_val = curr_obj_val

            if self.log:
                self.logmm.append([curr_obj_val, rel_err, err])
            if self.verbose:
                print(k, curr_obj_val, rel_err, err, N_continue)

            if rel_err <= self.stopThr[0] and err <= self.stopThr[1]:
                if (k - last_k) == 1:
                    N_continue += 1
                else:
                    N_continue = 1
                last_k = k
            else:
                N_continue = 0
            if k>self.min_iter and N_continue >= 10:
                self.n_total_iter = k
                break
        self.n_total_iter = k
        return (A_test, B_test, C_test)

if __name__ == "__main__":
    import sys
    sys.path.append("../")
    from simulate import generate_coupling_matrix, simu_data

    ## simu data
    r, N1, N2, I_len = 20, 123, 124, 1000                 # A, B, C
    m = generate_coupling_matrix(I_len, 400)              # TL, L
    n = generate_coupling_matrix(I_len, 300)              # TR, R
    K, epsilon = 1e4, 1                                   # M
    is_decom, d, solver = True, 10, "svd"                 # M1, M2

    print("### Simulation test")
    r, A, B, C, L, R, TL, TR, M, (M1, M2) = simu_data(r, N1, N2, I_len, m, n, K, epsilon, is_decom, d, solver)
    err = np.abs(M - np.dot(M1, M2.T))
    print("True data", r, A.shape, B.shape, C.shape, sep="\t")
    print("Simu data", L.shape, R.shape, TL.shape, TR.shape, M.shape, M1.shape, M2.shape)
    print("Error", err.sum(), err.max())
    #"""

    ## run: MMOTNTF
    mmot = MMOTNTD(L, R, M, TL, TR, epsilon=0.01, tau1=1, tau2=1, delta=1e-20, verbose=False)
    mmot.fit(r=20, init_params=[0, 0, 0], min_iter=0, max_iter=20)
    Ae, Be, Ce = mmot.factors
    res = mmot.calc_obj_res(Ae, Be, Ce)
    print( res )

    mmot = MMOTNTD(L, R, M, TL, TR, epsilon=0, tau1=1, tau2=1, delta=1e-20, verbose=False)
    mmot.fit(r=20, init_params=[0, 0, 0], min_iter=0, max_iter=20)
    Ae, Be, Ce = mmot.factors
    res = mmot.calc_obj_res(Ae, Be, Ce)
    print( res )

    mmot = MMOTNTD(L, R, (M1, M2), TL, TR, epsilon=0.01, tau1=1, tau2=1, delta=1e-20, verbose=False)
    mmot.fit(r=20, init_params=[0, 0, 0], min_iter=0, max_iter=20)
    Ae, Be, Ce = mmot.factors
    res = mmot.calc_obj_res(Ae, Be, Ce)
    print( res )

    mmot = MMOTNTD(L, R, (M1, M2), TL, TR, epsilon=0, tau1=1, tau2=1, delta=1e-20, verbose=False)
    mmot.fit(r=20, init_params=[0, 0, 0], min_iter=0, max_iter=20)
    Ae, Be, Ce = mmot.factors
    res = mmot.calc_obj_res(Ae, Be, Ce)
    print( res )




import numpy as np
from sklearn.decomposition import non_negative_factorization
from sklearn.utils.extmath import randomized_svd

def generate_coupling_matrix(n_row, n_col):
    """Generate coupling matrix for ligand or receptor

    Parameters
    ----------
    n_row : int
        Number of LR pairs
    n_col : int
        Number of ligand or receptor

    Returns
    -------
    (n_row, n_col) numpy.ndarray
        The binary matrix with at least non-zero element in each row and column.
        It records the composed relationship between LR pair and its ligand or receptor.
    """
    tmp1 = np.identity(n_col)
    tmp2 = np.random.random( size=(n_row-n_col, n_col) )
    tmp2 = (tmp2 >= tmp2.max(axis=1, keepdims=True)).astype(int)
    T = np.concatenate( (tmp1, tmp2), axis=0 )
    
    idx_row, idx_col = np.arange(n_row), np.arange(n_col)
    np.random.shuffle(idx_row)
    np.random.shuffle(idx_col)

    return T[idx_row, :][:, idx_col]

def simu_data(n_r=10, n1=123, n2=None, n_lr=500, 
              TL=None, TR=None,
              scale_factor=1e4, epsilon=1, 
              is_decomposition=False, n_d = 10, solver="nmf", eps=1e-100):
    """Generate systhesis data using inverse optimal transport (iOT)

    Parameters
    ----------
    n_r : int, optional
        Number of rank, by default 10
    n1 : int, optional
        Number of sender regions, by default 123
    n2 : None or int, optional
        Number of receiver regions, by default None
    n_lr : int, optional
        Number of LR pairs, by default 500
    TL : None, int or (n_lr, n_ligand) numpy.ndarray , optional
        The ligand coupling matrix (a binary matrix). 
        If None, it is a identity matrix; If int, it will generated a (n_lr, TL) binary matrix;
        If numpy.ndarray, use it for following generation. by default None
    TR : None, int or (n_lr, n_receptor) numpy.ndarray , optional
        The receptor coupling matrix (a binary matrix).
        If None, it is a identity matrix; If int, it will generated a (n_lr, TR) binary matrix;
        If numpy.ndarray, use it for following generation. by default None
    scale_factor : float or int, optional
        The factor to normalize the generated transport cost matrix M, by default 1e4
    epsilon : int, optional
        The coefficient for entropy regularization of generated transport cost matrix M, by default 1
    is_decomposition : bool, optional
        Whether to decompose transport cost matrix M, by default False
    n_d : int, optional
        Number of components, by default 10
    solver : str ("nmf" or "svd"), optional
        Numerical solver to perform matrix factorization of M, by default "nmf"
    eps : _type_, optional
        The coefficient to avoid zero in tensor factor A, B, C, by default 1e-100

    Returns
    -------
    n_r : int
        Number of rank. The same as input
    A : (n1, n_r) numpy.ndarray
        The generated sender loading matrix.
    B : (n2, n_r) numpy.ndarray
        The generated sender loading matrix.
    C : (n_lr, n_r) numpy.ndarray
        The generated sender loading matrix.
    L : (n1, n_ligand) numpy.ndarray
        The generated ligand expression matrix.
    R : (n2, n_receptor) numpy.ndarray
        The generated receptor expression matrix.
    TL : (n_lr, n_ligand) numpy.ndarray
        The generated ligand coupling matrix (a binary matrix).
    TR : (n_lr, n_receptor) numpy.ndarray
        The generated ligand coupling matrix (a binary matrix).
    M : (n1, n2) numpy.ndarray
        The transport cost matrix computed using inverse optimal transport.
    (M1, M2) : ( (n1, n_d) numpy.ndarray, (n2, n_d) numpy.ndarray ), optional
        The decomposed matrix from M. Only return when "is_decompostion is True".
    """
    ## check parameters
    if n2 is None:
        n2 = n1
    
    if TL is None:
        TL = np.identity(n=n_lr)
    elif isinstance(TL, int):
        assert TL <= n_lr, f"TL should less than or equal to n_lr, {TL}, {n_lr}"
        TL = generate_coupling_matrix(n_lr, TL)
    else:
        assert (TL.shape[0]==n_lr and TL.shape[1] <= n_lr), f"illegal data dimension, {TL.shape}, {n_lr}"
    
    if TR is None:
        TR = np.identity(n=n_lr)
    elif isinstance(TR, int):
        assert TR <= n_lr, f"TR should less than or equal to n_lr, {TR}, {n_lr}"
        TR = generate_coupling_matrix(n_lr, TR)
    else:
        assert (TR.shape[0]==n_lr and TR.shape[1] <= n_lr), f"illegal data dimension, {TR.shape}, {n_lr}"
    
    assert n_d <= max(n1, n2), "n_d should less than max{n1, n2}"

    ## true data
    A = np.random.random( size=(n1, n_r) )
    A /= A.sum(axis=0, keepdims=True)
    A = np.maximum(eps, A)
    B = np.random.random( size=(n2, n_r) )
    B /= B.sum(axis=0, keepdims=True)
    B = np.maximum(eps, B)
    C = np.maximum(eps, np.random.random( size=(n_lr, n_r) ))
    #print(A.shape, B.shape, C.shape)

    ## systhesis data
    L = np.dot(A * B.sum(axis=0, keepdims=True), np.dot(C.T, TL))
    R = np.dot(B * A.sum(axis=0, keepdims=True), np.dot(C.T, TR))

    M = np.exp(-np.dot(A * C.sum(axis=0, keepdims=True), B.T)/epsilon)
    M = M * (scale_factor / M.sum())
    #print(M.shape, np.isnan(M).sum() )

    if is_decomposition:
        if solver == "nmf":
            M1, M2, _ = non_negative_factorization(M, n_components=n_d, solver="cd", tol=1e-8, max_iter=5000)
            M2 = M2.T
        elif solver == "svd":
            tmp1, s, tmp2 = randomized_svd(M, n_components=n_d)
            s = np.sqrt(s).reshape(1, -1)
            M1 = tmp1 * s
            M2 = tmp2.T * s
            print("using svd")
        else:
            pass

        return n_r, A, B, C, L, R, TL, TR, M, (M1, M2)
    else:
        return n_r, A, B, C, L, R, TL, TR, M


if __name__ == "__main__":
    
    ################################
    ## simulation test1
    ################################
    
    r, N1, N2, I_len = 20, 123, 124, 1000      # A, B, C dimension
    m, n = 300, 400                            # TL, TR, L, R dimension
    K, epsilon = 1e4, 1                        # M
    is_decom, d, solver = True, 10, "nmf"      # M1, M2

    print("### Simulation test1")
    r, A, B, C, L, R, TL, TR, M, (M1, M2) = simu_data(r, N1, N2, I_len, m, n, K, epsilon, is_decom, d, solver)
    err = np.abs(M - np.dot(M1, M2.T))
    print("True data", r, A.shape, B.shape, C.shape, sep="\t")
    print("Simu data", L.shape, R.shape, TL.shape, TR.shape, M.shape, M1.shape, M2.shape)
    print("Error", err.sum(), err.max())

    ################################
    ## simulation test2
    ################################
    #"""
    r, N1, N2, I_len = 20, 123, 124, 1000      # A, B, C dimension
    m, n = None, 400                           # TL, TR, L, R dimension
    K, epsilon = 1e4, 1                        # M
    is_decom, d, solver = True, 10, "svd"      # M1, M2

    print("### Simulation test2")
    r, A, B, C, L, R, TL, TR, M, (M1, M2) = simu_data(r, N1, N2, I_len, m, n, K, epsilon, is_decom, d, solver)
    err = np.abs(M - np.dot(M1, M2.T))
    print("True data", r, A.shape, B.shape, C.shape, sep="\t")
    print("Simu data", L.shape, R.shape, TL.shape, TR.shape, M.shape, M1.shape, M2.shape)
    print("Error", err.sum(), err.max())
    #"""

    ################################
    ## simulation test3
    ################################
    #"""
    r, N1, N2, I_len = 20, 123, 124, 1000                 # A, B, C dimension
    m, n = None, generate_coupling_matrix(I_len, 200)     # TL, TR, L, R dimension
    K, epsilon = 1e4, 1                                   # M
    is_decom, d, solver = True, 10, "nmf"                 # M1, M2

    print("### Simulation test3")
    r, A, B, C, L, R, TL, TR, M, (M1, M2) = simu_data(r, N1, N2, I_len, m, n, K, epsilon, is_decom, d, solver)
    err = np.abs(M - np.dot(M1, M2.T))
    print("True data", r, A.shape, B.shape, C.shape, sep="\t")
    print("Simu data", L.shape, R.shape, TL.shape, TR.shape, M.shape, M1.shape, M2.shape)
    print("Error", err.sum(), err.max())
    #"""

    ################################
    ## simulation test4
    ################################
    #"""
    r, N1, N2, I_len = 20, 123, 124, 1000                 # A, B, C
    m = generate_coupling_matrix(I_len, 400)              # TL, L
    n = generate_coupling_matrix(I_len, 300)              # TR, R
    K, epsilon = 1e4, 1                                   # M
    is_decom, d, solver = True, 10, "svd"                 # M1, M2

    print("### Simulation test4")
    r, A, B, C, L, R, TL, TR, M, (M1, M2) = simu_data(r, N1, N2, I_len, m, n, K, epsilon, is_decom, d, solver)
    err = np.abs(M - np.dot(M1, M2.T))
    print("True data", r, A.shape, B.shape, C.shape, sep="\t")
    print("Simu data", L.shape, R.shape, TL.shape, TR.shape, M.shape, M1.shape, M2.shape)
    print("Error", err.sum(), err.max())
    #"""

    pass




import numpy as np
import pandas as pd
import scipy as sp

import plotly.express as px
import plotly.graph_objects as go
from pycirclize import Circos
from pycirclize.utils import ColorCycler

def plot_sankey(tca_obj, annot, loc=None, select_pattern=None, pad=8,
                title=None, save_name=None, is_show=True,
                font_size=16, figure_size=(800, 600)):
    """Sankey plot of communication patterns

    Parameters
    ----------
    tca_obj : TCA object 
        Constructed using A, B, C
    annot : pandas.DataFrame
        A regional annotation files
    loc : None or (N, 2) numpy.ndarray, optional
        The coordinate to plot the node. The number should be in[0, 1], by default None
    select_pattern : list, optional
        The index of selected patterns to plot. by default None
    pad : int, optional
        _description_, by default 8
    title : None or str, optional
        The name of the plot, by default None
    save_name : None or str, optional
        The path to save the figure, by default None
    is_show : bool, optional
        Whether to show the figure, by default True
    font_size : int, optional
        by default 16
    figure_size : tuple, optional
        by default (800, 600)
    """
    
    ## set node data
    nodeA = annot.copy()
    nodeB = annot.copy()
    nodeA["Pattern"] = "Pattern_"+pd.Series(tca_obj.A.argmax(axis=1)+1).astype(str)
    nodeB["Pattern"] = "Pattern_"+pd.Series(tca_obj.B.argmax(axis=1)+1).astype(str)

    ## set link data
    select_name = annot.columns.tolist()
    linkA = nodeA.value_counts(subset=select_name+["Pattern"], sort=False).reset_index()
    linkA.columns = select_name + ["Pattern", "Value"]
    linkB = nodeB.value_counts(subset=select_name+["Pattern"], sort=False).reset_index()
    linkB.columns = select_name + ["Pattern", "Value"]

    ## prepare data
    unique_select_name = annot[select_name[0]].unique().tolist()
    n_select_name = len(unique_select_name)
    if select_pattern is None:
        label_ = unique_select_name + [f"Pattern_{i}" for i in range(1, tca_obj.rank+1)] + unique_select_name
    else:
        label_ = unique_select_name + [f"Pattern_{i}" for i in select_pattern] + unique_select_name
    
    if loc is None:
        NODES = dict(label = label_, pad=pad)
    else:
        NODES = dict(label = label_, x = loc[:, 0], y = loc[:, 1], pad=pad)

    n_label = len(NODES["label"])
    n_add = n_label - n_select_name
    name2idx = dict( zip( NODES["label"][:-n_select_name], range(n_add) ) )
    LINKS = dict(
        source = linkA[select_name[0]].map(name2idx).values.tolist() + linkB["Pattern"].map(name2idx).values.tolist(),
        target = linkA["Pattern"].map(name2idx).values.tolist() + (linkB[select_name[0]].map(name2idx).values + n_add).tolist(),
        value  = linkA["Value"].values.tolist() + linkB["Value"].values.tolist()
    )

    ## plot
    data = go.Sankey(arrangement='freeform', node = NODES, link = LINKS)
    fig = go.Figure(data)
    if title is None:
        fig.update_layout(font_size=font_size, width=figure_size[0], height=figure_size[1] )
    else:
        fig.update_layout(title=title,
                          font_size=font_size, width=figure_size[0], height=figure_size[1])
    if save_name is not None:
        fig.write_image(save_name, format='pdf')
    if is_show:
        fig.show()
    pass


def get_lr_name_list(lr_list):
    l_names, r_names = set(), set()
    for tmp in lr_list:
        a, b = tmp.split('|', 1)
        l_names.add(a)
        r_names.add(b)
    return list(l_names), list(r_names)

def get_range(array):
    array_min, array_max = array.min(), array.max()
    array_min_int, array_max_int = int(array_min)-0.5, int(array_max)+0.5
    l = array_min_int if array_min_int <= array_min else array_min_int-0.5
    r = array_max_int if array_max_int >= array_max else array_max_int+0.5
    #if l<0 and r<0:
    #    r = 0
    #elif l>0 and r>0:
    #    l = 0
    return l, r

def plot_circos(tca_obj, annot, name, is_LR,
                strength_thr=0.0005, is_norm_expr=[True, True],
                sector_colors = ["#729FD3", "#E8C0A3", "#DAE1F2", "#A9CDB8", "#D8B8D7", "#C8C897", "#666666"],
                source_lobe = None, target_lobe = None,
                circ_lims = [50, 360], alpha_scale = 1, is_show = True, save_name = None
                ):
    ## 01. preprocessing: get network
    if is_LR:
        assert name in tca_obj.lr_anno.index, "The name of LR pair must be in the index of tca_obj.lr_anno."
        net = tca_obj.get_LR_network(idx=[name], return_LR_name=False).mean(axis=0)
        l_name, r_name = name.split("|")
        l_expr, r_expr = tca_obj.L[l_name].values, tca_obj.R[r_name].values
    else:
        assert name in tca_obj.lr_anno["pathway"].values, "The name of LR pair must be in the tca_obj.lr_anno['pathway']."
        net, lr_names = tca_obj.get_pathway_network(pathway_name=name, return_LR_name=True)
        net = net.mean(axis=0)
        l_names, r_names = get_lr_name_list(lr_names)
        l_expr, r_expr = tca_obj.L[l_names].mean(axis=1).values, tca_obj.R[r_names].mean(axis=1).values

    net /= net.max()
    if not (is_norm_expr is None):
        if is_norm_expr[0]:
            l_expr = sp.stats.zscore(l_expr)
        if is_norm_expr[1]:
            r_expr = sp.stats.zscore(r_expr)

    ## 02. preprocessing: get annotation
    Lobe = annot["Lobe"].str.split(" ", expand=True, n=1)[0].drop_duplicates().values
    lobe_reindex = np.concatenate([ annot.loc[annot["Lobe"].str.find(i)>=0].copy().reset_index().index.values for i in Lobe])
    assert len(Lobe) == len(sector_colors), "The number of lobe must match the number of given colors in 'sector colors' "

    ## 03. plot
    sectors = { i:(annot["Lobe"].str.find(i) >= 0).sum() for i in Lobe }
    sector_colors = { i:j for i, j in zip(Lobe, sector_colors) }

    circos = Circos(sectors, space=2, start=circ_lims[0], end=circ_lims[1],)
    for sector in circos.sectors:
        #sector.axis(fc="none", ls="dashdot", lw=2, ec="black", alpha=0.5)
        #sector.text(f"{sector.name} ({sector.size})", r=60, size=10)
        tmp_idx = annot["Lobe"].str.find(sector.name)>=0

        ## 01. name label
        lobe_track = sector.add_track( r_lim=(65, 72) )
        lobe_track.text(sector.name, size=12)
        lobe_track.axis(fc=sector_colors[sector.name])

        ## 03. receptor
        x, y = np.arange(sector.start, sector.end)+0.5, r_expr[tmp_idx]
        tmp_range = get_range(r_expr[tmp_idx])
        receptor_track = sector.add_track( r_lim=(74, 86) )
        receptor_track.bar(x, y, vmin=tmp_range[0], vmax=tmp_range[1], width=0.5)
        receptor_track.line(x, y, vmin=tmp_range[0], vmax=tmp_range[1], lw=1, color="gray")
        receptor_track.line(np.arange(sector.start, sector.end+1), [0]*(len(x)+1), 
                        vmin=tmp_range[0], vmax=tmp_range[1], lw=1.5, color="black")
        receptor_track.axis()
        
        ## 04. ligand expression
        x, y = np.arange(sector.start, sector.end)+0.5, l_expr[tmp_idx]
        tmp_range = get_range(l_expr[tmp_idx])
        ligand_track = sector.add_track( r_lim=(88, 100) )
        ligand_track.bar(x, y, vmin=tmp_range[0], vmax=tmp_range[1], width=0.5)
        ligand_track.line(x, y, vmin=tmp_range[0], vmax=tmp_range[1], lw=1, color="gray")
        ligand_track.line(np.arange(sector.start, sector.end+1), [0]*(len(x)+1), 
                        vmin=tmp_range[0], vmax=tmp_range[1], lw=1.5, color="black")
        ligand_track.axis()

        ## 05 add links
        tmp_net = net[tmp_idx, :]
        for i in range(tmp_net.shape[0]):
            if source_lobe is None or sector.name in source_lobe:
                param1 = (sector.name, i, i+1)
            else:
                continue          
            for j in range(tmp_net.shape[1]):
                j_name = annot["Lobe"].iloc[j].split(" ")[0]
                if target_lobe is None or j_name in target_lobe:
                    param2 = (j_name, lobe_reindex[j], lobe_reindex[j]+1)
                else:
                    continue
                if tmp_net[i, j] < strength_thr:
                    continue
                circos.link(param1, param2, color=sector_colors[sector.name], 
                            direction=1, alpha=tmp_net[i, j]*alpha_scale, r1=65, r2=65)

    text_common_kws = dict(ha="left", va="center", size=12)
    circos.text("Ligand expression", r=94, color="black", **text_common_kws)
    circos.text("Receptor expression", r=80, color="black", **text_common_kws)
    circos.text("Lobe", r=68.5, color="black", **text_common_kws)
    circos.text(name, size=15, r=105)
    if is_show:
        fig = circos.plotfig(figsize=(8, 8))
    if save_name is not None:
        circos.savefig(save_name)
    pass




import numpy as np
import pandas as pd

import os

class LRdatabase(object):
    """ Extract LR regional expression and thier relationship from gene expression and LR database

    Parameters
    ----------
    lrdb: None, str or pandas.DataFrame
        if None, use built-in database which were based on CellChat and NeuroChat.
        else input your own curated database with at least 5 columns in order: (tsv format)
            "interaction_id", index column, 
            "lignad symbol": genes related to ligand, seperated by "_" if multiple gene to a lignad
            "receptor symbol", genes related to ligand, seperated by "_" if multiple gene to a lignad
            "lr_type", see lr_type for example
            "is_neurontransmitter", bool variable, if this ligand receptor interaction related to neurontransmitter.
    lr_type: list, choose from "Secreted", "ECM-Receptor", "Cell-Cell Contact", "Non-protein Signaling" or their combination.
    level: str, choose from "lr" or "gene". 
        "lr": extract ligand and receptor expression by geometric mean;
        "gene": means extract the gene expression related to ligand and receptors
    
    Attributes
    ----------
    L : (n_sender, n_ligand) pandas.DataFrame
        The ligand expression matrix.
    R : (n_receiver, n_receptor) pandas.DataFrame
        The receptor expression matrix.
    TL : (n_lr, n_ligand) pandas.DataFrame
        The generated ligand coupling matrix.
    TR : (n_lr, n_receptor) pandas.DataFrame
        The generated ligand coupling matrix.    
    """
    def __init__(self, lrdb=None, lr_type=["Secreted", "Non-protein Signaling"], level="lr"):
        if lrdb is None:
            curr_dir = os.path.dirname(__file__)
            lrdb1 = pd.read_csv(os.path.join(curr_dir, "./LRdatabase/lrdb_cell.tsv"), sep="\t", index_col=0)
            lrdb1["is_neurontransmitter"] = False
            lrdb2 = pd.read_csv(os.path.join(curr_dir, "./LRdatabase/lrdb_neuron.tsv"), sep="\t", index_col=0)
            lrdb2["is_neurontransmitter"] = True
            lrdb = pd.concat([lrdb1, lrdb2])
        elif isinstance(lrdb, str):
            lrdb = pd.read_csv(lrdb)
        else:
            pass
        assert lrdb.shape[1]>=5, "at least 5 columns for lrdb"

        self.lrdb = lrdb.loc[lrdb.iloc[:, 2].isin(lr_type), :]
        self.level = level
    
    def extract_lr_expression(self, expr, level=None, method="geo"):
        """Extract regional expression matrix of ligand and receptor.

        Parameters
        ----------
        expr : str or (n_region, n_genes) pandas.DateFrame
            The regional gene expression matrix
        level : None or str, optional
            Construct the coupling matrix at "gene" level or "lr" level, by default None
        method : str, optional
            The method to get ligand and receptor expression from genes, by default "geo"
        """
        if level is not None:
            self.level = level

        ## 00. load expression data
        expr = self._load_expr_data(expr)
        expr_colsum = expr.sum(axis=0)
        expr = expr.loc[:, expr_colsum>0]

        if level is not None:
            self.level = level

        if self.level == "lr":
            L, R = self._extract_expr_level_lr(expr, method=method)
        elif self.level == "gene":
            L, R = self._extract_expr_level_gene(expr)
        else:
            pass
        self.L = L
        self.R = R

    def generate_coupling_matrix(self, L=None, R=None, level=None):
        if level is not None:
            self.level = level
        if L is not None:
            self.L = L
        if R is not None:
            self.R = R

        TL, TR = {}, {}
        if self.level == "lr":
            for lr_id in self.lrdb.index:
                l_id, r_id = lr_id.split("|")
                if l_id in self.L.columns and r_id in self.R.columns:
                    TL[lr_id] = {l_id:1}
                    TR[lr_id] = {r_id:1}
    
        elif self.level == "gene":
            for lr_id in self.lrdb.index:
                l_id, r_id = lr_id.split("|")
                l_sym, r_sym = l_id.split("_"), r_id.split("_")
                l_isin = [ g in self.L.columns for g in l_sym ]
                r_isin = [ g in self.R.columns for g in r_sym ]
                if np.alltrue(l_isin) and np.alltrue(r_isin):
                    TL[lr_id] = { g:1 for g in l_sym}
                    TR[lr_id] = { r:1 for r in r_sym}
        else:
            pass

        ## fill nan with 0
        TL = pd.DataFrame(TL).T
        TL[np.isnan(TL)] = 0
        TR = pd.DataFrame(TR).T
        TR[np.isnan(TR)] = 0
        self.TL = TL
        self.TR = TR

    def _extract_expr_level_lr(self, expr, method):
        L, R = {}, {}
        for l_sym, r_sym in self.lrdb.values[:, :2]:
            l_syms, r_syms = l_sym.split("_"), r_sym.split("_")
            l_isin = [ l in expr.columns for l in l_syms ]
            r_isin = [ r in expr.columns for r in r_syms ]
            if (not np.alltrue(l_isin)) or (not np.alltrue(r_isin)):
                continue
            
            ## geometric mean
            if method == "geo":
                if l_sym not in L:
                    L[l_sym] = np.power( np.prod(expr.loc[:, l_syms].values, axis=1), 1/len(l_syms) )
                if r_sym not in R:
                    R[r_sym] = np.power( np.prod(expr.loc[:, r_syms].values, axis=1), 1/len(r_syms) )
            ## mean
            elif method == "mean":
                if l_sym not in L:
                    L[l_sym] = np.mean(expr.loc[:, l_syms].values, axis=1)
                if r_sym not in R:
                    R[r_sym] = np.mean(expr.loc[:, r_syms].values, axis=1)
            ## min
            elif method == "min":
                if l_sym not in L:
                    L[l_sym] = np.min(expr.loc[:, l_syms].values, axis=1)
                if r_sym not in R:
                    R[r_sym] = np.min(expr.loc[:, r_syms].values, axis=1)
            else:
                pass

        L = pd.DataFrame(L)
        L.index = expr.index
        R = pd.DataFrame(R)
        R.index = expr.index
        return L, R

    def _extract_expr_level_gene(self, expr):
        L, R = {}, {}
        for l_sym, r_sym in self.lrdb.values[:, :2]:
            l_syms, r_syms = l_sym.split("_"), r_sym.split("_")
            l_isin = [ l in expr.columns for l in l_syms ]
            r_isin = [ r in expr.columns for r in r_syms ]
            if (not np.alltrue(l_isin)) or (not np.alltrue(r_isin)):
                continue

            for g in l_syms:
                if g not in L:
                    L[g] = expr.loc[:, g].values
            for g in r_syms:
                if g not in R:
                    R[g] = expr.loc[:, g].values


        L = pd.DataFrame(L)
        L.index = expr.index
        R = pd.DataFrame(R)
        R.index = expr.index
        return L, R
    
    def _load_expr_data(self, expr):
        if isinstance(expr, str):
            if expr.endswith("parq"):
                return pd.read_parquet(expr)
            else:
                return pd.read_csv(expr)
        else:
            return expr


if __name__ == "__main__":
    import sys

    ## test1
    filename = sys.argv[1]
    level = "lr"
    lrdb_obj = LRdatabase(level=level)
    lrdb_obj.extract_lr_expression(filename)
    lrdb_obj.generate_coupling_matrix()
    print(lrdb_obj.L.shape, lrdb_obj.R.shape, lrdb_obj.TL.shape, lrdb_obj.TR.shape)

    ## test2
    filename = pd.read_parquet(sys.argv[1])
    level = "lr"
    lrdb_obj = LRdatabase(level=level)
    lrdb_obj.extract_lr_expression(filename)
    lrdb_obj.generate_coupling_matrix()
    print(lrdb_obj.L.shape, lrdb_obj.R.shape, lrdb_obj.TL.shape, lrdb_obj.TR.shape)



from setuptools import setup, find_packages, find_namespace_packages

from pathlib import Path
this_directory = Path(__file__).parent
README = (this_directory / "README.md").read_text()

setup(
    name='clria',
    version='1.0.7',
    author='DU Zongchang',
    license='MIT',
    url='https://github.com/duzc-Repos/CLRIA',
    
    description='Package to decipher LRI-mediated brain network communication',
    long_description=README,
    long_description_content_type='text/markdown',
    classifiers=[
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
    ],
    
    #include_package_data=True,
    package_data = {
        '':['*.txt', '*.xlsx', '*.tsv', '*.csv'],
        'preprocessing.LRdatabase':['*.tsv']
    },
    #packages=find_packages(exclude=['CLRIA_tutorial', 'tests*']),
    packages=find_namespace_packages(where="clria"),
    package_dir={"": "clria"},
    
    install_requires=[
        'numpy',  'pandas', 'tensorly',
        'scipy', 'statsmodels',
        'scikit_learn',
        'nibabel', 'netneurotools',
        'POT',
        'plotly',
    ],
    python_requires='>=3.8',
    
)



import numpy as np
import ot
from scipy.optimize import fixed_point

class OTRegression(object):

    def __init__(self, pattern, regP=1, lambda_=2) -> None:
        """
        @parm pattern: list of matrix with equal size.
        @parm regP: int or float, regular parameters for Sinkhorm algorithm
        @parm lambda_: int or float, ridge regular parameters for $\beta$
        """
        self.X = pattern.copy()
        self.rank = len(self.X)
        self.regP = regP
        self.lambda_ = lambda_
        pass

    def _Entropy(self, P):
        tmp = P[P>0]
        return -(tmp * (np.log(tmp) - 1)).sum()
    def _squre_norm(self, beta):
        return (beta ** 2).sum()
    
    def get_M(self, beta):
        return np.exp( sum([ -beta[i] * self.X[i] for i in range(self.rank) ]) )

    def get_opt_obj_val(self, beta, P):
        M = self.get_M(beta)
        return (M * P).sum() - self.regP * self._Entropy(P) + self.lambda_/2 * self._squre_norm(beta)
    
    def get_opt_OT_dist(self, beta, P):
        M = self.get_M(beta)
        return (M * P).sum()
    
    def _check_src_dst(self, src, dst):
        src = np.array(src)
        dst = np.array(dst)
        assert np.allclose(src.sum(), dst.sum(), rtol=1e-8, atol=1e-8), f"sum(src) != sum(dst), {src.sum()}, {dst.sum()}"
        return src, dst

    def fit(self, src, dst, max_iter=(500, 1000, 500), delta=1e-16, min_iter=10, log=False):
        """
        @param src: array or list, source state or source active map
        @param dst: array or list, target state or target active map
        @param max_iter: tuple or list, 
        """

        def contractive_mapping(beta, lambda_, X, P):
            """
            @param beta: array
            @param lambda_: int or float
            #param X: list of matrix with equal size
            """
            rank = len(X)
            M = np.exp( sum([ -beta[i] * X[i] for i in range(rank) ]) )
            mid = P * M
            return np.array([ (X[i] * mid).sum() for i in range(rank) ]) / lambda_

        ## 00. Initialization
        src, dst = self._check_src_dst(src, dst)
        beta = np.ones(shape=self.rank) / self.rank

        ## 01. Iteration
        for k in range(max_iter[0]):
            
            ## Update P
            M = self.get_M(beta)
            P = ot.sinkhorn(a=src, b=dst, M=M, reg=self.regP, method="sinkhorn_log", numItermax=max_iter[1], log=False)
            #P = ot.emd(src, dst, M=M, numItermax=max_iter[1])

            ## Update beta
            beta = fixed_point(contractive_mapping, beta, args=(self.lambda_, self.X, P), maxiter=max_iter[2])

            ## Criterion
            if k == 0:
                dist_prev = self.get_opt_OT_dist(beta, P)
                continue
            
            dist_curr = self.get_opt_OT_dist(beta, P) * 1e8
            rel_err = np.abs(dist_curr - dist_prev) / (dist_prev)
            if log:
                print(k, dist_curr, rel_err, sep="\t")
            if k < min_iter or rel_err > delta:
                dist_prev = dist_curr
            else:
                return beta, P
        
        msg = f"Failed to converge after {max_iter} iterations, value is {dist_curr}, rel_err is {rel_err}"
        raise RuntimeError(msg)

class UOTRegression(object):

    def __init__(self, pattern, regP=1, regm=1, lambda_=2) -> None:
        """
        @parm pattern: list of matrix with equal size.
        @parm regP: int or float, regular parameters for Sinkhorm algorithm
        @parm regm: int or float, regular parameters for mass constraints
        @parm lambda_: int or float, ridge regular parameters for $\beta$
        """
        self.X = pattern.copy()
        self.rank = len(self.X)
        self.regP = regP
        self.regm = regm
        self.lambda_ = lambda_
        pass

    def _Entropy(self, P):
        tmp = P[P>0]
        return -(tmp * (np.log(tmp) - 1)).sum()
    def _squre_norm(self, beta):
        return (beta ** 2).sum()
    def _KL_div(self, x, y):
        return (x * (np.log(x) - np.log(y) - 1)).sum() + y.sum()
    
    def get_M(self, beta):
        return np.exp( sum([ -beta[i] * self.X[i] for i in range(self.rank) ]) )
    def get_opt_UOT_dist(self, beta, P):
        M = self.get_M(beta)
        return (M * P).sum()
    def get_opt_obj_val(self, beta, P, src, dst):
        M = self.get_M(beta)
        kl_term = self.regm * self._KL_div(P.sum(axis=0), src) + self.regm * self._KL_div(P.sum(axis=1), dst)
        return (M * P).sum() - self.regP * self._Entropy(P) + kl_term + self.lambda_/2 * self._squre_norm(beta)
    
    def _check_src_dst(self, src, dst):
        src = np.array(src)
        dst = np.array(dst)
        return src, dst

    def fit(self, src, dst, beta_init=None, max_iter=(500, 1000, 500), delta=1e-16, min_iter=10, log=False):
        """
        @param src: array or list, source state or source active map
        @param dst: array or list, target state or target active map
        @beta_init: None, str or numpy.ndarray, initial beta value for iteration
        @param max_iter: tuple or list, 
        """

        def contractive_mapping(beta, lambda_, X, P):
            """
            @param beta: array
            @param lambda_: int or float
            #param X: list of matrix with equal size
            """
            rank = len(X)
            M = np.exp( sum([ -beta[i] * X[i] for i in range(rank) ]) )
            mid = P * M
            return np.array([ (X[i] * mid).sum() for i in range(rank) ]) / lambda_

        ## 00. Initialization
        src, dst = self._check_src_dst(src, dst)
        if beta_init is None:
            beta = np.ones(shape=self.rank) / self.rank
        elif beta_init == "random":
            beta = np.random.random(size=self.rank)
        else:
            beta = beta_init.copy()

        ## 01. Iteration
        for k in range(max_iter[0]):
            
            ## Update P
            M = self.get_M(beta)
            P = ot.sinkhorn_unbalanced(a=src, b=dst, M=M, reg=self.regP, reg_m=self.regm, 
                                       method="sinkhorn", numItermax=max_iter[1], log=False)
            #P = ot.emd(src, dst, M=M, numItermax=max_iter[1])

            ## Update beta
            beta = fixed_point(contractive_mapping, beta, args=(self.lambda_, self.X, P), maxiter=max_iter[2])

            ## Criterion
            if k == 0:
                dist_prev = self.get_opt_UOT_dist(beta, P)
                continue
            
            dist_curr = self.get_opt_UOT_dist(beta, P) * 1e8
            rel_err = np.abs(dist_curr - dist_prev) / (dist_prev)
            if log:
                print(k, dist_curr, rel_err, sep="\t")
            if k < min_iter or rel_err > delta:
                dist_prev = dist_curr
            else:
                return beta, P
        
        msg = f"Failed to converge after {max_iter} iterations, value is {dist_curr}, rel_err is {rel_err}"
        raise RuntimeError(msg)
